==> oth-client-0.11.0.jar: doccla/oth_client/clinician/api/patients/patients.clj <==
(ns doccla.oth-client.clinician.api.patients.patients
  (:require
   [clj-http.client :as client]
   [doccla.oth-client.schemas :as schemas]
   [doccla.oth-client.utils :as utils]
   [malli.clj-kondo :as clj-kondo]
   [malli.core :as m]
   [malli.instrument :as mi]
   [malli.util :as mu]))

(def patient-group-schema
  [:map [:name string?] [:links [:map [:patient-group string?]]]])

(def relative-schema-base
  [:map
   {:closed true}
   [:first-name string?]
   [:last-name string?]
   [:relation string?]
   [:phone string?]
   [:address {:optional true} string?]
   [:city {:optional true} string?]
   [:note {:optional true} string?]
   [:email {:optional true} string?]
   [:guardian {:optional true}
    [:map
     [:username string?]
     [:email {:optional true} string?]
     [:unique-id {:optional true} string?]]]])

(def relative-schema-read
  (mu/merge
   relative-schema-base
   [:map
    [:unique-id {:optional true} string?]
    [:username {:optional true} string?]
    [:links [:map [:relative string?]]]]))

(def relative-schema-create relative-schema-base)

(def patient-schema-base
  [:map
   {:closed true}
   [:unique-id string?]
   [:first-name string?]
   [:last-name string?]
   [:address string?]
   [:postal-code string?]
   [:city string?]
   [:place {:optional true} string?]
   [:phone {:optional true} string?]
   [:mobile-phone {:optional true} string?]
   [:email {:optional true} string?]
   [:profile-picture {:optional true} [:re schemas/base-64-regex]]
   [:sex [:enum "male" "female" "unknown"]]
   [:due-date {:optional true} [:re schemas/simple-date-regex]]
   [:username string?]
   [:date-of-birth {:optional true} [:re schemas/simple-date-regex]]
   [:comment {:optional true} string?]])

(def patient-status-schema
  [:enum "active" "deceased" "discharged" "discharged_self_managed" "paused"])

(def patient-schema-read
  (mu/merge
   patient-schema-base
   [:map
    [:created-date string?]
    [:status patient-status-schema]
    [:relatives [:vector relative-schema-read]]
    [:patient-groups
     [:vector patient-group-schema]]
    [:links
     [:map
      [:measurements string?]
      [:questionnaire-results string?]
      [:contact-info string?]
      [:questionnaire-schedules string?]
      [:notifications string?]
      [:thresholds string?]
      [:questionnaires string?]
      [:self string?]
      [:change-password string?]]]]))

(m/=> get-patient [:=>
                   [:cat schemas/opts-schema [:or :string :int]]
                   [:or schemas/error-schema (schemas/success-schema patient-schema-read)]])

(defn get-patient-id-by-url [url]
  (second (re-matches #".*/patients/(\d+)" url)))

(defn ^:mockable get-patient
  "Retrieve a specific patient."
  [opts id]
  (let [res (client/get (str (:base-url opts) "/clinician/api/patients/" id)
                        (utils/opts->request opts))
        post-processor (fn [data] (let [f (if (:validate-output? opts) m/coerce m/decode)]
                                    (f patient-schema-read data utils/prune-map-transformer)))]
    (utils/->output [200] post-processor res)))

;;;
;;; Create patient
;;; 

(def patient-schema-create
  (mu/merge patient-schema-base
            [:map
             [:relatives [:vector relative-schema-create]]
             [:links [:map
                      {:closed true}
                      [:patient-groups  [:vector {:min 1}
                                         [:re {:error/message "Must be a valid URL"} schemas/url-regex]]]
                      [:data-responsible {:optional true}
                       [:re {:error/message "Must be a valid URL"} schemas/url-regex]]]]]))

(def patient-schema-create-resp
  (mu/merge patient-schema-read
            [:map
             [:relatives
              [:vector (mu/assoc relative-schema-read
                                 [:temporary-password {:optional true}] string?)]]
             [:temporary-password {:optional true} string?]]))

(m/=> create-patient [:=>
                      [:cat schemas/opts-schema patient-schema-create]
                      [:or schemas/error-schema (schemas/success-schema patient-schema-create-resp)]])

(defn ^:mockable create-patient
  "Create a new patient."
  [opts patient]
  (let [res (client/post (str (:base-url opts) "/clinician/api/patients/")
                         (merge (utils/opts->request opts)
                                {:body (utils/encode patient)}))
        post-processor (fn [data] (let [f (if (:validate-output? opts) m/coerce m/decode)]
                                    (f patient-schema-create-resp data utils/prune-map-transformer)))]
    (utils/->output [201] post-processor res)))

(def patient-schema-update
  (-> patient-schema-create
      (mu/assoc :status patient-status-schema)
      (mu/optional-keys)))

(defn insert-updates
  "Given a patient response (i.e. a map conforming to patent-schema-read)
   and a set of updates (i.e. a map conforming to patient-schema-update)
   returns a new map that satisfies the patient/{id}/PUT endpoint."
  [patient updates]
  (-> patient
      ;; move the patient groups to links
      ;; (the get enpoint returns a map of patient groups, but
      ;; the put endpoint expects a list of links) 
      (assoc :links {:patient-groups (vec (map (comp :patient-group :links)
                                               (:patient-groups patient)))})
      (dissoc :patient-groups)
      ;; remove any other un-needed keys for update
      (dissoc :created-date)
      ;; then merge in any updates. 
      (merge updates)))

(m/=> update-patient [:=>
                      [:cat schemas/opts-schema [:or :string :int] patient-schema-update]
                      [:or schemas/error-schema (schemas/success-schema patient-schema-create-resp)]])
(defn ^:mockable update-patient
  "Update an existing patient. Differs slightly from the OTH API in that
   it supports partial updates rather than wholesale replacement of every value.
   N.B. Patient groups and relatives will be replaced wholesale if provided."
  [opts id updates]
  (let [{:keys [success? response] :as get-resp} (get-patient opts id)]
    (if success?
      (let [_ (tap> response)
            res (client/put (str (:base-url opts) "/clinician/api/patients/" id)
                            (merge (utils/opts->request opts)
                                   {:body (-> (insert-updates response updates)
                                              (utils/encode))}))
            post-processor (fn [data] (let [f (if (:validate-output? opts) m/coerce m/decode)]
                                        (f patient-schema-create-resp data utils/prune-map-transformer)))]
        (utils/->output [200] post-processor res))

      ;; If the get-patient step failed, return that failure as the response
      get-resp)))

;;;
;;; Common
;;;

;; Enable instrumentation so library users get schema checking.
(mi/instrument! {:filters [(-> *ns* str symbol mi/-filter-ns)]
                 :scope #{:input}})
(clj-kondo/emit!)
;; Enable mocks
(utils/make-mockable)

(comment
  (require '[clojure.edn :as edn])
  (def creds (edn/read-string (slurp "creds/oth.edn")))
  (def opts {:base-url    "https://doccla-dev.oth.io"
             :validate-output? true
             :auth {:type :id-secret :id (:id creds) :secret (:secret creds)}})

  (get-patient (utils/->mock-client {'get-patient "foo"}) 1)
  (def p1 (:response (get-patient opts 1)))

  (create-patient
   opts
   {:first-name  "John"
    :last-name   "Doe"
    :username    "JohnDoe102"
    :sex         "male"
    :postal-code "123"
    :address     "abc"
    :city        "ny"
    :links       {:patient-groups
                  ["https://doccla-dev.oth.io/clinician/api/patientgroups/27"]}
    :relatives   []
    :unique-id   "JohnDoe102"})

  (update-patient opts 1 {:sex "male"}))

==> oth-client-0.17.0.jar: doccla/oth_client/clinician/api/patients/patients.clj <==
(ns doccla.oth-client.clinician.api.patients.patients
  (:require
   [clj-http.client :as client]
   [doccla.oth-client.schemas :as schemas]
   [doccla.oth-client.utils :as utils]
   [malli.clj-kondo :as clj-kondo]
   [malli.core :as m]
   [malli.instrument :as mi]
   [malli.util :as mu]))

(def patient-group-schema
  [:map [:name string?] [:links [:map [:patient-group string?]]]])

(def relative-schema-base
  [:map
   {:closed true}
   [:first-name string?]
   [:last-name string?]
   [:relation string?]
   [:phone string?]
   [:address {:optional true} string?]
   [:city {:optional true} string?]
   [:note {:optional true} string?]
   [:email {:optional true} string?]
   [:guardian {:optional true}
    [:map
     [:username string?]
     [:email {:optional true} string?]
     [:unique-id {:optional true} string?]]]])

(def relative-schema-read
  (mu/merge
   relative-schema-base
   [:map
    [:unique-id {:optional true} string?]
    [:username {:optional true} string?]
    [:links [:map [:relative string?]]]]))

(def relative-schema-create relative-schema-base)

(def patient-schema-base
  [:map
   {:closed true}
   [:unique-id string?]
   [:first-name string?]
   [:last-name string?]
   [:address string?]
   [:postal-code string?]
   [:city string?]
   [:place {:optional true} string?]
   [:phone {:optional true} string?]
   [:mobile-phone {:optional true} string?]
   [:email {:optional true} string?]
   [:profile-picture {:optional true} [:re schemas/base-64-regex]]
   [:sex [:enum "male" "female" "unknown"]]
   [:due-date {:optional true} [:re schemas/simple-date-regex]]
   [:username string?]
   [:date-of-birth {:optional true} [:re schemas/simple-date-regex]]
   [:comment {:optional true} string?]])

(def patient-status-schema
  [:enum "active" "deceased" "discharged" "discharged_self_managed" "paused"])

(def patient-schema-read
  (mu/merge
   patient-schema-base
   [:map
    [:created-date string?]
    [:status patient-status-schema]
    [:relatives [:vector relative-schema-read]]
    [:patient-groups
     [:vector patient-group-schema]]
    [:links
     [:map
      [:measurements string?]
      [:questionnaire-results string?]
      [:contact-info string?]
      [:questionnaire-schedules string?]
      [:notifications string?]
      [:thresholds string?]
      [:questionnaires string?]
      [:self string?]
      [:change-password string?]]]]))

(m/=> get-patient [:=>
                   [:cat schemas/opts-schema [:or :string :int]]
                   [:or schemas/error-schema (schemas/success-schema patient-schema-read)]])

(defn get-patient-id-by-url [url]
  (second (re-matches #".*/patients/(\d+)" url)))

(defn ^:mockable get-patient
  "Retrieve a specific patient."
  [opts id]
  (let [res (client/get (str (:base-url opts) "/clinician/api/patients/" id)
                        (utils/opts->request opts))
        post-processor (fn [data] (let [f (if (:validate-output? opts) m/coerce m/decode)]
                                    (f patient-schema-read data utils/prune-map-transformer)))]
    (utils/->output [200] post-processor res)))

;;;
;;; Create patient
;;; 

(def patient-schema-create
  (mu/merge patient-schema-base
            [:map
             [:relatives [:vector relative-schema-create]]
             [:links [:map
                      {:closed true}
                      [:patient-groups  [:vector {:min 1}
                                         [:re {:error/message "Must be a valid URL"} schemas/url-regex]]]
                      [:data-responsible {:optional true}
                       [:re {:error/message "Must be a valid URL"} schemas/url-regex]]]]]))

(def patient-schema-create-resp
  (mu/merge patient-schema-read
            [:map
             [:relatives
              [:vector (mu/assoc relative-schema-read
                                 [:temporary-password {:optional true}] string?)]]
             [:temporary-password {:optional true} string?]]))

(m/=> create-patient [:=>
                      [:cat schemas/opts-schema patient-schema-create]
                      [:or schemas/error-schema (schemas/success-schema patient-schema-create-resp)]])

(defn ^:mockable create-patient
  "Create a new patient."
  [opts patient]
  (let [res (client/post (str (:base-url opts) "/clinician/api/patients/")
                         (merge (utils/opts->request opts)
                                {:body (utils/encode patient)}))
        post-processor (fn [data] (let [f (if (:validate-output? opts) m/coerce m/decode)]
                                    (f patient-schema-create-resp data utils/prune-map-transformer)))]
    (utils/->output [201] post-processor res)))

(def patient-schema-update
  (-> patient-schema-create
      (mu/assoc :status patient-status-schema)
      (mu/optional-keys)))

(defn insert-updates
  "Given a patient response (i.e. a map conforming to patent-schema-read)
   and a set of updates (i.e. a map conforming to patient-schema-update)
   returns a new map that satisfies the patient/{id}/PUT endpoint."
  [patient updates]
  (-> patient
      ;; move the patient groups to links
      ;; (the get enpoint returns a map of patient groups, but
      ;; the put endpoint expects a list of links) 
      (assoc :links {:patient-groups (vec (map (comp :patient-group :links)
                                               (:patient-groups patient)))})
      (dissoc :patient-groups)
      ;; remove any other un-needed keys for update
      (dissoc :created-date)
      ;; then merge in any updates. 
      (merge updates)))

(m/=> update-patient [:=>
                      [:cat schemas/opts-schema [:or :string :int] patient-schema-update]
                      [:or schemas/error-schema (schemas/success-schema patient-schema-create-resp)]])
(defn ^:mockable update-patient
  "Update an existing patient. Differs slightly from the OTH API in that
   it supports partial updates rather than wholesale replacement of every value.
   N.B. Patient groups and relatives will be replaced wholesale if provided."
  [opts id updates]
  (let [{:keys [success? response] :as get-resp} (get-patient opts id)]
    (if success?
      (let [_ (tap> response)
            res (client/put (str (:base-url opts) "/clinician/api/patients/" id)
                            (merge (utils/opts->request opts)
                                   {:body (-> (insert-updates response updates)
                                              (utils/encode))}))
            post-processor (fn [data] (let [f (if (:validate-output? opts) m/coerce m/decode)]
                                        (f patient-schema-create-resp data utils/prune-map-transformer)))]
        (utils/->output [200] post-processor res))

      ;; If the get-patient step failed, return that failure as the response
      get-resp)))

;;; get patients

;;; The following schema was auto-generated
;;; from the openapi schema

(def list-patients-schema
  [:schema
   {:registry
    {:patient-definitions/patientStatus
     [:and [:string {}] [:enum "active" "deceased" "discharged" "discharged_self_managed" "paused"]],
     :patient-definitions/patient
     [:map
      [:address {:optional true} [:string {}]]
      [:email {:optional true} :any]
      [:profilePicture {:optional true} [:or [:not :string] [:re "^data:image.+;base64,"]]]
      [:sex {:optional true} [:and [:string {}] [:enum "male" "female" "unknown"]]]
      [:place {:optional true} :any]
      [:phone {:optional true} :any]
      [:city {:optional true} [:string {}]]
      [:username {:optional true} [:string {}]]
      [:mobile-phone {:optional true} :any]
      [:first-name {:optional true} [:string {}]]
      [:date-of-birth
       {:optional true}
       [:or nil? [:re "^(\\d{4})[\\-](0[1-9]|1[0-2])[\\-]([12]\\d|0[1-9]|3[01])$"]]]
      [:due-date {:optional true} [:string {}]]
      [:last-name {:optional true} [:string {}]]
      [:comment {:optional true} :any]
      [:postal-code {:optional true} [:string {}]]
      [:unique-id {:optional true} [:string {}]]],
     :definitions/linkType [:re "^https?:\\/\\/[-\\w]+(\\.[-\\w]+)*(:[0-9]+)?\\/?(\\/[-.\\w]*)*.*$"]}}
   [:map
    [:offset [:int {}]]
    [:total [:int {}]]
    [:max [:int {}]]
    [:links
     [:map
      [:next {:optional true} :definitions/linkType]
      [:previous {:optional true} :definitions/linkType]
      [:self :definitions/linkType]]]
    [:results
     [:cat
      [:*
       [:and
        :patient-definitions/patient
        [:map
         [:status {:optional true} :patient-definitions/patientStatus]
         [:created-date {:optional true} [:string {}]]
         [:links [:map [:patient :definitions/linkType]]]]]]
      [:* :any]]]]])

(def query-params-schema
  [:map
   {:closed true}
   [:uniqueId {:optional true} string?]
   [:username {:optional true} string?]
   [:firstName {:optional true} string?]
   [:lastName {:optional true} string?]
   [:search {:optional true} string?]
   [:phone {:optional true} string?]
   [:patientGroup {:optional true} [:or string?
                                    [:sequential {:max 100} string?]]]
   [:id {:optional true} [:or string?
                          [:sequential {:max 1000} string?]]]
   [:status {:optional true} string?]
   [:select {:optional true} string?]

   [:offset {:optional true} int?]
   [:max {:optional true} int?]])

(m/=> get-patients [:=>
                    [:cat schemas/opts-schema query-params-schema]
                    [:or schemas/error-schema (schemas/success-schema list-patients-schema)]])

(defn ^:mockable get-patients
  "Retrieve a specific patient."
  [opts params]
  (let [res (client/get (str (:base-url opts) "/clinician/api/patients/")
                        (assoc (utils/opts->request opts)
                               :query-params params))
        post-processor (fn [data] (let [f (if (:validate-output? opts) m/coerce m/decode)]
                                    (f list-patients-schema data utils/prune-map-transformer)))]
    (utils/->output [200] post-processor res)))

;;;
;;; Common
;;;

;; Enable instrumentation so library users get schema checking.
(mi/instrument! {:filters [(-> *ns* str symbol mi/-filter-ns)]
                 :scope #{:input}})
(clj-kondo/emit!)
;; Enable mocks
(utils/make-mockable)

(comment
  (require '[clojure.edn :as edn])
  (def creds (edn/read-string (slurp "creds/oth.edn")))
  (def opts {:base-url    "https://doccla-dev.oth.io"
             :validate-output? true
             :auth {:type :id-secret :id (:id creds) :secret (:secret creds)}})

  (get-patient (utils/->mock-client {'get-patient "foo"}) 1)
  (def p1 (:response (get-patient opts 1)))

  (get-patients opts {:firstName "jo" :max 2})

  (create-patient
   opts
   {:first-name  "John"
    :last-name   "Doe"
    :username    "JohnDoe102"
    :sex         "male"
    :postal-code "123"
    :address     "abc"
    :city        "ny"
    :links       {:patient-groups
                  ["https://doccla-dev.oth.io/clinician/api/patientgroups/27"]}
    :relatives   []
    :unique-id   "JohnDoe102"})

  (update-patient opts 1 {:sex "male"}))
